import time
import urllib
import logging
import http

import itertools as it
import pandas as pd

from Bio import Entrez
from typing import Callable, Any, Union, Iterable


def get_chunklimits(iterable_length: int, chunksize: int) -> tuple[int, int]:
    """
    takes the length of a given iterable and the size of the chunks we want to create
    and returns tuples of min and max idx for each of these chunks

    :param iterable_length:    length of the iterable to chunkify
    :param chunksize:          size of the iterable chunks

    :yield:                    a tuple of min idx and max idx
    """
    retmin, retmax = 0, 0
    for i in range(0, iterable_length, chunksize):
        retmin = i 
        retmax = i + chunksize
        yield retmin, retmax


def create_webenv(uids: list[Union[int, str]], db: str) -> dict[str, str]:
    """
    posts the given list of uids to the Entrez history server with the given db
    and returns query key and web env id

    :param uids:   list of uids
    :param db:     string denoting the database the UIDs are belonging to

    :return:       dictionary with 'QueryKey' and 'WebEnv' keys containig the query key and web env ID
    """
    response_handle = retry(
        Entrez.epost,
        db = db,
        id = ','.join(str(uid) for uid in uids)
    )
    return Entrez.read(response_handle)


def print_exception_and_increment_try(n_tries: int, exception: Exception, sleep: int = 10) -> int:
    """
    prints the raised exception, sleeps for sleep seconds, increments the trial counter by one and returns it
    
    :param n_tries:     current number of tries
    :param exception:   raised exception object
    :param sleep:       number of seconds to wait before retrying
    
    :return:            current number of tries incremented by one
    """
    logging.warn(exception)
    logging.warn(f'current try: {n_tries}')
    time.sleep(sleep)
    n_tries += 1
    return n_tries


def retry(func: Callable, *args, n_retries: int = 5, sleep: int = 10, **kwargs) -> Any:
    """
    takes a function that might fail with one of the following exceptions: HTTPError, RuntimeError or IncompleteRead
    and retries n_retries times before raising a RuntimeError. Sleeps sleep seconds before retry to resolve possible
    too many request errors with remote APIs
    
    :param func:         function that might fail and needs to be retried
    :param *args:        any positional arguments passed to func
    :param n_retries:    number of retries before raising RuntimeError
    :param sleep:        seconds to sleep before retry
    :param **kwargs:     any keyword arguments passed to func
    
    :return:             any result func might yield
    """
    n_tries = 1
    success = False
    while not success:
        if n_tries == n_retries:
            raise RuntimeError('Exceeded retries!')
            
        try:
            result = func(*args, **kwargs)
            success = True
            
        # this can also be handled in principle via the Entrez.max_tries
        # and Entrez.sleep_between_tries but we leave this as is for now
        # see https://biopython.org/docs/dev/api/Bio.Entrez.html
        except urllib.error.HTTPError as e:
            n_tries = print_exception_and_increment_try(n_tries, e, sleep)
        
        # this exception seems to be raised when the search backend times out
        # due to too many requests I suspect so we retry and hope for the best
        except RuntimeError as e:
            n_tries = print_exception_and_increment_try(n_tries, e, sleep)
        
        # sometimes happens when reading the response if so we just retry
        except http.client.IncompleteRead as e:
            n_tries = print_exception_and_increment_try(n_tries, e, sleep)
            
        
    return result


def process_in_chunks(items: Iterable[Any], processing_function: Callable, chunksize = 50000, **kwargs) -> list[Any]:
    """
    processes the given iterable with processing_function in chunks of size chunksize

    :param items:                   iterable containing objects that need to be processed with processing_function
    :param processing_function:     function that takes the iterable and processes it additional arguments can be given with **kwargs
    :param chunksize:               number of items per processing chunk

    :return:                        list of whatever processing_function returns for each chunk    
    """
    chunks = it.batched(
        items,
        n = chunksize
    )

    n = 0
    results = []
    for chunk in chunks:
        logging.info(f'retrieving items {n} to {n + len(chunk)}')
        n += len(chunk)

        result = processing_function(chunk, **kwargs)

        results.append(result)

    return results
        


import h5py
import gc
import scipy

import pandas as pd
import numpy as np
import anndata as ad

from os import PathLike
from . import concurrency
from typing import Union, Iterable, Any


def parse_table(archs4_file: Union[str, PathLike], root_key: str, table_key: str) -> dict[str, Any]:
    """
    parses the archs4 h5 file and extracts the data at /root_key/table_key into a dictionary
    see https://maayanlab.cloud/archs4/help.html for more infos on the available options of
    root_key and table_key
    
    :param archs4_file:   path to the archs4 h5 file to parse
    :param root_key:      root key to use for parsing the file metadata
    :param table_key:     table key to use for parsing the file metadata
    
    :return:              dictionary containing the metadata with column names as keys and column contents as values
    """
    archs4 = h5py.File(archs4_file, 'r')
    
    data = dict()
    table = archs4[root_key][table_key]
    for key in table.keys():
        column_data_piece = table[key][0]
        if isinstance(column_data_piece, np.number):
            column_data = table[key][:]
        
        else:
            column_data = [x.decode('utf-8') for x in table[key][:]]
        
        data[key] = column_data
    
    return data


def filter_table_data(data: dict, retain_keys: Iterable[str]) -> dict[str, Any]:
    """
    filters the data dictionary to retain only the keys specified by retain_keys
    
    :param data:          dictionary as returned by parse_table
    :param retain_keys:   list of data keys to retain
    
    :return:              filtered data dictionary
    """
    retain_data = {
        key: data[key].copy() for key in retain_keys
    }
    return retain_data


def get_srx_samn_from_items(relation_items: Iterable[str]) -> tuple[str, str]:
    """
    takes list of strings generated from an item in the 'relation' column
    of the /meta/samples table in the archs4 h5 file and extracts the SRA and
    BioSample accessions from it
    
    :param relation_items:   list of strings of the format 'key: value'
    
    :return:                 sra and bisample accession as strings
    """
    srx, samn = None, None
    for item in relation_items:
        if not item:
            continue
            
        key, value = item.split(': ')
            
        if key == 'SRA':
            srx = value.split('=')[-1]
        
        if key == 'BioSample':
            samn = value.split('/')[-1]
        
    return srx, samn


def extract_srx_and_samn_accessions(relation_data: Iterable[str]) -> dict[str, list[str]]:
    """
    takes the full 'relation' column of the archs4 h5 /meta/samples table and extracts
    SRA and BioSample accessions for each of the items contained
    
    :param relation_data:  list of strings of the format 'key1: value1, key2: value2, ...'
    
    :return:               dictionary containing keys srx_accession, biosample_accession with lists of these accessions as values
    """
    srx_samn_accessions = {
        key: [] for key in ['srx_accession', 'biosample_accession']
    }
    for relation in relation_data:
        relation_items = relation.split(',')
        srx, samn = get_srx_samn_from_items(
            relation_items
        )
        srx_samn_accessions['srx_accession'].append(srx)
        srx_samn_accessions['biosample_accession'].append(samn)
    
    return srx_samn_accessions


def get_filtered_sample_metadata(archs4_file: Union[str, PathLike], keys_to_retain: Iterable[str]) -> pd.DataFrame:
    """
    parses the archs4 h5 file and returns a pandas.DataFrame of with all columns of the /meta/samples table 
    specified in keys_to_retain plus extracted SRA and BioSamples accessions (columns srx_accession and biosample_accession)
    
    :param archs4_file:       path to the archs4 file to parse
    :param keys_to_retain:    list of keys in the /meta/samples table to retain
    
    :return:                  pandas.DataFrame containing the respective metadata + SRA and BioSample accessions
    """
    data = parse_table(
        archs4_file,
        'meta',
        'samples'
    )

    filtered_data = filter_table_data(
        data, 
        keys_to_retain
    )

    srx_samn_accessions = extract_srx_and_samn_accessions(
        filtered_data['relation']
    )

    srx_samn_table = pd.DataFrame().from_dict(
        srx_samn_accessions,
        orient = 'columns'
    )

    table = pd.DataFrame().from_dict(
        filtered_data, 
        orient = 'columns'
    )

    table = pd.concat(
        [table, srx_samn_table],
        axis = 1
    )

    del data, filtered_data, srx_samn_accessions
    gc.collect()
    
    return table


# adapted from archs4py as this would not install due to issues with Python 3.12 and numpy requirement
def index(file, sample_idx, gene_idx = [], silent=False, n_processes = 1):
    """
    Retrieve gene expression data from a specified file for the given sample and gene indices.

    Args:
        file (str): The file path or object containing the data.
        sample_idx (list): A list of sample indices to retrieve expression data for.
        gene_idx (list, optional): A list of gene indices to retrieve expression data for. Defaults to an empty list (return all).
        silent (bool, optional): Whether to disable progress bar. Defaults to False.

    Returns:
        pd.DataFrame: A pandas DataFrame containing the gene expression data.
    """
    sample_idx = sorted(sample_idx)
    gene_idx = sorted(gene_idx)
    row_encoding = get_encoding(file)
    with h5py.File(file, "r") as f:
        genes = np.array([x.decode("UTF-8") for x in np.array(f[row_encoding])])
        if len(sample_idx) == 0:
            return pd.DataFrame(index=genes[gene_idx])
        gsm_ids = np.array([x.decode("UTF-8") for x in np.array(f["meta/samples/geo_accession"])])[sample_idx]

    if len(gene_idx) == 0:
        gene_idx = list(range(len(genes)))

    data = concurrency.process_data_in_chunks(
        zip(sample_idx, gsm_ids),
        get_samples,
        n_processes=n_processes,
        file = file,
        gene_idx = gene_idx
    )

    exp = ad.AnnData(
        X = scipy.sparse.vstack(data),
        var = pd.DataFrame(index = genes[gene_idx]),
        obs = pd.DataFrame(index = gsm_ids)
    )
    exp.var_names_make_unique()
    exp.eliminate_zeros()
    return exp


def consecutive(
    sequence_array: np.ndarray, 
    *additional_arrays_to_split: np.ndarray,  
    stepsize: int = 1
) -> list[np.ndarray]:
    """
    finds all sequences of consecutive elements in a numpy array
    and splits the input arrays accordingly
    """
    split_idx = np.where(np.diff(sequence_array) != stepsize)[0]+1
    arrays = [sequence_array, *additional_arrays_to_split]
    return [np.split(array, split_idx) for array in arrays]


def get_encoding(file):
    with h5py.File(file) as f:
        if "genes" in list(f["meta"].keys()):
            if "gene_symbol" in list(f["meta/genes"].keys()):
                return "meta/genes/gene_symbol"
            elif "symbol" in list(f["meta/genes"].keys()):
                return "meta/genes/symbol"
        elif "transcripts" in list(f["meta"].keys()):
            if "ensembl_id" in list(f["meta/trancripts"].keys()):
                return "meta/trancripts/ensembl_id"
        else:
            raise Exception("error in gene/transcript meta data")


def read_sample_data(file, idx, gsm, gene_idx):
    with h5py.File(file, "r") as f:
        dense_expression = f["data/expression"][:, idx][gene_idx, :]
        sparse_expression = scipy.sparse.csr_matrix(dense_expression.T)
        sparse_expression.eliminate_zeros()

        del dense_expression
        gc.collect()

    return sparse_expression
    

# this is a wrapper to ensure compliance with API
def get_samples(sample_idx_gsms, file, gene_idx, **kwargs):
    sample_idxs = np.array([item[0] for item in sample_idx_gsms])
    sample_gsms = np.array([item[1] for item in sample_idx_gsms])
        
    data = [
        read_sample_data(file, consecutive_idx, consecutive_gsm, gene_idx) 
        for consecutive_idx, consecutive_gsm in zip(*consecutive(sample_idxs, sample_gsms))
    ]
    return scipy.sparse.vstack(data)


def samples(file, sample_ids, silent=False, n_processes = 1):
    sample_ids = set(sample_ids)
    with h5py.File(file, "r") as f:
        samples = [x.decode("UTF-8") for x in np.array(f["meta/samples/geo_accession"])]

    idx = [i for i,x in enumerate(samples) if x in sample_ids]
    if len(idx) > 0:
        return index(file, idx, silent=silent, n_processes=n_processes)
    


import logging
import certifi
import urllib3

import pandas as pd

from . import parsers
from . import summary
from . import concurrency
from . import dbutils
from . import metasra
from typing import Union, Iterable
from os import PathLike
from io import BytesIO

logging.basicConfig(
    format = '%(threadName)s: %(asctime)s-%(levelname)s-%(message)s',
    datefmt = '%Y-%m-%d %H:%M:%S',
    level = logging.INFO
)

def map_accessions_to_srauids(
    table: pd.DataFrame, 
    outfilename: Union[PathLike, str], 
    chunksize: int = 5000, 
    n_processes: int = 1
) -> None:
    """
    takes a pandas.DataFrame containing columns srx_accession, biosample_accession and geo_accession, maps those accessions
    to SRA UIDs and writes the resulting map to outfilename. If n_processes > 1 this will be done concurrently
    
    :param table:         pandas.DataFrame containing columns srx_accession, biosample_accession and geo_accession
    :param db:            string denoting the NCBI database to retrieve the UIDs from
    :param outfilename:   path to the outputfile the resulting map should be written to
    :param chunksize:     size of the indiviually processed chunks of the input table
    :param n_processes:   number of processes to use for mapping if n_processes > 1 this will be done concurrently using multiprocessing.imap
    
    :return:              None
    """
    mapping_table = parsers.get_not_yet_mapped(
        table,
        outfilename
    )
    
    concurrency.process_data_in_chunks(
        mapping_table.iterrows(),
        dbutils.map_accessions_to_uids,
        db = 'sra',
        outfilename = outfilename
    )


def determine_accession_type(accession: str) -> str:
    """
    determines the type of accession used for retrieving the UIDs
    currently identifies SRX SAMN or GSM accessions

    :param accession:   string containing the accession to be identified

    :return:            string giving the type of accession
    """
    types = {
        'experiment': 'SRX',
        'biosample': 'SAMN',
        'geo': 'GSM'
    }
    
    for accession_type, accession_prefix in types.items():
        if accession.startswith(accession_prefix):
            return accession_type
        

def merge_uids_and_accessions(uids: pd.DataFrame, accessions: pd.DataFrame) -> pd.DataFrame:
    """
    merges the uid frame with the retrieved accessions

    :param uids:            pandas.DataFrame containing 'uid' and 'accession' columns
    :param accessions:      pandas.DataFrame containing accessions correspondig to uids

    :return:                pandas.DataFrame merged on uid
    """
    uids = uids.copy()
    uids['accession_type'] = uids.accession.apply(
        determine_accession_type
    )

    merged_groups = []
    for accession_type, group in uids.groupby('accession_type'):
        group = group.rename(
            columns = {'accession': accession_type}
        )
        
        merged_group = group.merge(
            accessions,
            on = accession_type,
            how = 'inner'
        )

        merged_groups.append(merged_group)
        
    return pd.concat(merged_groups)


def srauids_to_accessions(srauids: pd.DataFrame, chunksize: int = 50000) -> pd.DataFrame:
    """
    retrieves all available accessions for the sra uids given in 'uid' column of srauids

    :param srauids:     pandas.DataFrame containing an 'accession' and a 'uid' column
    :param chunksize:   integer giving the size of the chunks posted and retrieved from Entrez

    :return:            pandas.DataFrame containing experiment, sample, biosample, study, bioproject and geo accessions
    """
    accessions = summary.summaries_from_uids(
        srauids.uid,
        'sra',
        parsers.accessions_from_esummary_response,
        parsers.accession_matchers,
        chunksize = chunksize
    )
    
    return merge_uids_and_accessions(srauids, accessions)


def biosample_uids_to_metadata(biosample_uids: Iterable[Union[str, int]], chunksize: int = 50000) -> pd.DataFrame:
    """
    retrieves all available metadata for the biosample uids

    :param biosample_uids:  iterable of biosample UIDs as string or int
    :param chunksize:       integer giving the size of the chunks posted and retrieved from Entrez

    :return:            pandas.DataFrame containing experiment, sample, biosample, study, bioproject and geo accessions
    """
    metadata = summary.summaries_from_uids(
        biosample_uids,
        'biosample',
        parsers.metadata_from_biosample_uids,
        parsers.biosample_metadata_parsers,
        chunksize = chunksize
    )
    
    return metadata


def metasra_from_study_id(study_ids: Iterable[str]) -> pd.DataFrame:
    """
    used the MetaSRA API to retrieve normalized metdata for all samples in 
    a study given by study_ids from their database
    
    :param study_ids:   iterable containing SRA Biostudy accessions

    :return:            pandas.DataFrame containing MetaSRA normalised metadata
    """
    metasra_data = dbutils.process_in_chunks(
        study_ids,
        metasra.study_id_to_metasra,
        chunksize = 1000
    )

    return pd.concat(metasra_data)


def normalize_string(string: str) -> str:
    """
    removes non ascii characters from a given string

    :param string:      string from which to remove non ascii characters

    :return:            string with just ascii characters
    """
    normalized_string = []
    for char in string:
        if ord(char) < 128 and char != ' ':
            normalized_string.append(char)
    
    return ''.join(normalized_string)


def all_equal(x: pd.Series) -> bool:
    """
    checks if all items in a pandas.Series are the same

    :param x:       pandas.Series object to check

    :return:        True if all items are the same else False
    """
    ref = normalize_string(x.iloc[0])
    return all(normalize_string(item) == ref for _, item in x.items())


def merge_to_annotated_metadata_frame(
    biosample_metadata: pd.DataFrame, 
    srauids_to_biosampleuids: pd.DataFrame,
    ncbi_accessions: pd.DataFrame,
    archs4_metadata: pd.DataFrame
) -> pd.DataFrame:
    """
    combines the passed pandas.DataFrames to a fully annotated ARCHS4 metadata set

    :param biosample_metadata:          pandas.DataFrame containing BioSample metadata (see biosample_uids_to_metadata)
    :param srauids_to_biosampleuids:    pandas.DataFrame containing the linked SRA to BioSample UIDs (see link.link_sra_to_biosample)
    :param ncbi_accessions:             pandas.DataFrame containing the SRA accessions (see srauids_to_accessions)
    :param archs4_metadata:             pandas.DataFrame containing the extracted ARCHS4 metadata (see archs4.get_filtered_sample_metadata)

    :return:                            pandas.DataFrame containing annotated ARCHS4 metadata
    """
    biosample_uids_and_metadata = srauids_to_biosampleuids.merge(
        biosample_metadata,
        on = 'biosample',
        how = 'inner'
    )

    accessions_biosample_metadata_uids = biosample_uids_and_metadata \
        .rename(
            columns = {
                'sra': 'sra_uid',
                'biosample': 'biosample_uid',
                'title': 'biosample_title',
                'attribute': 'raw_biosample_metadata'
            }
        ).merge(
            ncbi_accessions.rename(
                columns = {
                    'uid': 'sra_uid'
                }
            ),
            on = 'sra_uid',
            how = 'inner'
        )

    archs4_annotated = accessions_biosample_metadata_uids.merge(
        archs4_metadata.drop(
            columns = [
                    'srx_accession', 
                    'biosample_accession'
                ]
        ).rename(
            columns = {
                'geo_accession': 'geo',
                'title': 'geo_title',
                'source_name_ch1': 'geo_source_name',
                'characteristics_ch1': 'geo_metadata'
            }
        ),
        on = 'geo',
        how = 'inner'
    )

    # this is done in order to remove scrambled annotations due to spurious links to BioSample
    spurious_duplication = archs4_annotated.sra_uid.duplicated(keep = False)
    archs4_annotated_nonduplicated = archs4_annotated.loc[~spurious_duplication, :]
    archs4_annotated_duplicated = archs4_annotated.loc[spurious_duplication, :]

    all_equal_idx = archs4_annotated_duplicated[['geo_title', 'biosample_title']].apply(
        all_equal,
        axis = 1
    )

    archs4_annotated_deduplicated = pd.concat(
        [
            archs4_annotated_nonduplicated,
            archs4_annotated_duplicated.loc[all_equal_idx, :]
        ]
    )
    return archs4_annotated_deduplicated.reset_index(drop = True)
